#include <AccelStepper.h>
#include <MultiStepper.h>


//Neck Stepper
AccelStepper Neck(1,40,41);
int neckPos = 0;
int neckCenteringDist = 250;
int neckSensor;
int neckPin = 22;
bool isCentered = false;

//Drive Steppers
AccelStepper RightFrontWheel(1, 20, 21); // Stepper1
AccelStepper RightBackWheel(1, 44, 45);   // Stepper2
AccelStepper LeftFrontWheel(1, 18, 19);  // Stepper3
AccelStepper LeftBackWheel(1, 42, 43);  // Stepper4
MultiStepper steppers;
long positions[4] = {0,0,0,0};
int wheelSpeed = 2000;

//BrushStepper
AccelStepper Brush(AccelStepper::FULL4WIRE,26,28,27,29);
int brushUP = 500;
int brushDOWN = 0;

//Sensors
int ESpin_fl = 37;
int ESpin_fr = 35;
int ESpin_bl = 36;
int ESpin_br = 34;

int ES_fl; 
int ES_fr;
int ES_bl;
int ES_br;

bool edgeDetected = false;
int ReboundAngle = 0;

int IRpin = 25;

//Other
float MovementCommand [4] = {0,0,0,0};
bool Run = false;
int Draw = 0;
int Counter = 0;
int Distance = 0;
int CounterLimit;
float CounterCompFactor = 0.4;
int Angle = 0;


void setup() 
{
  //Neck Stepper
  Neck.setMaxSpeed(200);
  Neck.setAcceleration(10000); 
  pinMode(neckPin,INPUT);

  //Brush Stepper
  Brush.setMaxSpeed(500);
  Brush.setAcceleration(10000);

  //Drive Steppers
  LeftFrontWheel.setMaxSpeed(wheelSpeed);
  RightFrontWheel.setMaxSpeed(wheelSpeed);
  RightBackWheel.setMaxSpeed(wheelSpeed);
  LeftBackWheel.setMaxSpeed(wheelSpeed);
  steppers.addStepper(LeftFrontWheel);
  steppers.addStepper(LeftBackWheel);
  steppers.addStepper(RightFrontWheel);
  steppers.addStepper(RightBackWheel);

  //Sensors
  pinMode(ESpin_fl,INPUT);
  pinMode(ESpin_fr,INPUT);
  pinMode(ESpin_bl,INPUT);
  pinMode(ESpin_br,INPUT);
  
  //Other
  Serial.begin(115200);

  
}
 
void loop() 
{

//  if (isCentered == false)
//  {
//    HomeNeck();
//  }

  Neck.setMaxSpeed(2000);
  
  Run = true;
  
  Angle = random(0,360);
  Distance = random(100,500);
  Draw = random(0,100);
  PrepairMove(Angle,Distance,Draw);
  
  while (Run)
  {
    if (steppers.run() == 0)
    {
      CalculateMotions(MovementCommand);
      Counter++;
    }
    if (Counter >= CounterLimit)
    {
      Run = false;
      Counter = 0;
    }
    CheckEdgeSensors();
    if (edgeDetected)
    {
      Run = false;
      Counter = 0;
      ReboundAngle = abs((Angle + 180)%360);
      positions[0] = 0;
      positions[1] = 0;
      positions[2] = 0;
      positions[3] = 0;
    }
  }

  if (edgeDetected)
  {
    Angle = ReboundAngle + random(-30,30);
    if (Angle < 0)
    {
      Angle = 360 + Angle;
    }
    Distance = random(100,500);
    Draw = random(0,100);
    PrepairMove(Angle,Distance,Draw);
    while (edgeDetected)
    {
      if (steppers.run() == 0)
      {
        CalculateMotions(MovementCommand);
        Counter++;
      }
      if (Counter >= CounterLimit)
      {
        edgeDetected = false;
        Counter = 0;
      }
    }
  }
}

///////////////////////////////////////////////////////////////////

void CheckEdgeSensors()
{
  ES_fl = digitalRead(ESpin_fl);
  ES_fr = digitalRead(ESpin_fr);
  ES_bl = digitalRead(ESpin_bl);
  ES_br = digitalRead(ESpin_br);
  if (ES_fl or ES_fr or ES_bl or ES_br)
  {
    edgeDetected = true;
  }
}


///////////////////////////////////////////////////////////////////

void MoveBrush(int target)
{
  Brush.moveTo(target);
  while (Brush.run() == 1){}
}

///////////////////////////////////////////////////////////////////

void PrepairMove(int InAngle,int InDistance,int InDraw)
{
  int EvalAngle = abs(InAngle%360);
  int AngleMod = EvalAngle%90;

//Move Neck
  int neckTarget = EvalAngle;
  if (neckTarget <= 180)
  {
    MoveNeck(neckTarget);
  }
  else
  {
    neckTarget = 360 - neckTarget;
    MoveNeck(-neckTarget);
  }

//Eval Angle
  if (0 <=EvalAngle && EvalAngle < 90)
  {
    MovementCommand[0] = -45;
    MovementCommand[1] = AngleMod - 45;
    MovementCommand[2] = -AngleMod + 45;
    MovementCommand[3] = 45;
  }
  else if (EvalAngle < 180)
  {
    MovementCommand[0] = AngleMod - 45;
    MovementCommand[1] = 45;
    MovementCommand[2] = -45;
    MovementCommand[3] = -AngleMod + 45;
  }
  else if (EvalAngle < 270)
  {
    MovementCommand[0] = 45;
    MovementCommand[1] = -AngleMod + 45;
    MovementCommand[2] = AngleMod - 45;
    MovementCommand[3] = -45;
  }
  else
  {
    MovementCommand[0] = -AngleMod + 45;
    MovementCommand[1] = -45;
    MovementCommand[2] = 45;
    MovementCommand[3] = AngleMod - 45;
  }

  //Modify CounterLimit for distance
  if (AngleMod > 45)
  {
    AngleMod = 90 - AngleMod;
  }
  float tempFactor = InDistance/1.75;
  int distFactor = tempFactor;
  float angleFactor = distFactor * ((sqrt(sq(45)-sq(AngleMod-45))/45)*CounterCompFactor);
  CounterLimit = distFactor + angleFactor;

  //Move Pen
  if (InDraw >= 50)
  {
    MoveBrush(brushDOWN);
  }
  else
  {
    MoveBrush(brushUP);
  }
}

///////////////////////////////////////////////////////////////////

void CalculateMotions(float Command[4]) 
{
  positions[0] += Command[0];
  positions[1] += Command[1];
  positions[2] += Command[2];
  positions[3] += Command[3];
  steppers.moveTo(positions);
}

///////////////////////////////////////////////////////////////////

void HomeNeck()
{
  neckSensor = digitalRead(neckPin);
  isCentered = false;
  if (neckSensor == 1)
  {
    isCentered = true;
  }
  else
  {
    Neck.moveTo(neckCenteringDist);
    while (Neck.run() == 1)
    {
      neckSensor = digitalRead(neckPin);
      if (neckSensor == 1)
      {
        isCentered = true;
        break;
      }
    }
    if (isCentered != true);
    {
      Neck.moveTo(-neckCenteringDist);
      while (Neck.run() == 1)
      {
        neckSensor = digitalRead(neckPin);
        if (neckSensor == 1)
        {
          isCentered = true;
          break;
        }
      }
    } 
  }
  if (isCentered)
  {
    Neck.setCurrentPosition(0);
    Serial.println("Neck Centered");
  }
  else
  {
    Serial.println("Neck Center Not Found");
  }
}

///////////////////////////////////////////////////////////////////

void MoveNeck(float targetAngle)
{
  float tempTarget = (targetAngle/360)*3200;
  int target = tempTarget;
  Neck.moveTo(target);
  while (Neck.run() == 1){}
}
